function [Xff,Pff,Xfp,Pfp,FInn,LogLK_est,Xf0_est,Theta_est,PFInn,mpts_out]=InnovEstimators(Xf0Index,ThetaIndex,EstimOptions)
%    Innovation estimates of the unknown parameters theta and the unobserved componet x of the 
% continuous model
%        dx = f(t,x,theta) dt + g(t,x,theta) dw, 
% with dicrete observation equation 
%         Z = h(t,x,theta) + p(t,x,theta)n + e, 
% given a set of discrete observation of Z. The initial filter value Xf0 can be also estimated.
%
% Imputs
%   Xf0Index:     Indexes of the Xf0 to be estimates ([] is no one)
%   ThetaIndex:   Indexes of the theta to be estimates ([] is no one)
%   EstimOptions: Full infomation about the Space State Model to be estimated
%
% OutPuts:
%   Xff : LL fitting-filter  
%   Pff : LL fitting-filter variance 
%   Xfp : LL fitting-prediction 
%   Pfp : LL fitting-prediction variance  
%   FInn: fitting-Innovation
%   LogLK_est:  -2log(Likelihood)
%   Xf0_est:   innovation estimates of Xf0
%   Theta_est: innovation estimates of Theta
%   PFInn: fitting-Innovation variance
%
% An call to the function estimset.m MUST be done before to use InnovEstimators.m
%

%  References
%   - "Approximate continuous-discrete filters for the estimation of diffusion processes 
%      from partial and noisy observations", 
%     Jimenez J.C. http://arxiv.org/abs/1212.3721.
%
%  See also next reference for problems with complete observations
%   - "Approximate discrete-time schemes for the estimation of diffusion processes
%     from complete observations", 
%     Jimenez J.C.  http://arxiv.org/abs/1212.1788.  
%
%  See also next reference for a review of inference methods for SDEs
%   - "Inference methods for discretely observed continuous-time stochastic volatility models: 
%     a commented overview,
%     Jimenez J.C., Biscay R., Ozaki T., Asia-Pacific Financial Markets, 12 (2006) 109-141.
%

% The Likelihood function and the parameters to be estimated are normalized according the good optimization
% practice sugested at Chapter "Practicalities" in:
% Gill P.E., Murray W. and Wright M.H., Practical Optimization. Academic Press, 1981.

Xf00  = EstimOptions{3};
Theta0= EstimOptions{5};
LB = EstimOptions{10};
UB = EstimOptions{11};
OptimOptions = EstimOptions{12};
OptimMethod  = EstimOptions{17};

EstimOptions{13} = EstimChoise(ThetaIndex,Xf0Index);
EstimOptions{14} = Xf0Index;
EstimOptions{15} = ThetaIndex;

switch EstimOptions{13},
   case 0, 
     n_par  = length(Xf0Index) + length(ThetaIndex);
     ParVar = [Xf00(Xf0Index)' Theta0(ThetaIndex)];
   case 1, 
     n_par  = length(ThetaIndex);
     ParVar = Theta0(ThetaIndex);
   case 2, 
     n_par  = length(Xf0Index);
     ParVar = Xf00(Xf0Index)';
end
NoZeroParVar = find(abs(ParVar)>eps);
ZeroParVar   = find(abs(ParVar)<eps);
  
if ~(isempty(LB) & isempty(UB)),
  WrongLB=find(LB>ParVar);
  if ~isempty(WrongLB), disp('A lower bound is wrong. Please channge the bounds # : '), WrongLB; error(' '), end 
  WrongUB=find(UB<ParVar);
  if ~isempty(WrongUB), disp('A upper bound is wrong. Please channge the bounds # : '), WrongUB; error(' '), end 
  LB(NoZeroParVar) = LB(NoZeroParVar)./ParVar(NoZeroParVar);
  LB(ZeroParVar)   = LB(ZeroParVar) + 1;
  UB(NoZeroParVar) = UB(NoZeroParVar)./ParVar(NoZeroParVar);
  UB(ZeroParVar)   = UB(ZeroParVar) + 1;
  index=find(LB>UB);
  if ~isempty(index)
    temp=LB;
    LB(index)=UB(index);
    UB(index)=temp(index);
  end
end

figure
ParVar = ones(1,n_par);   
objfun = @(ParVar) LogLKFunction(ParVar,EstimOptions);
if isempty(EstimOptions{9}),
  confun = [];
else     
  confun = @(ParVar) ConFunLike(ParVar,EstimOptions);
end;    
warning('off','all');
switch OptimMethod
  case 'fmincon'
    [Par_est, F_val, ExitFlag] = fmincon(objfun,ParVar,[],[],[],[],LB,UB,confun,OptimOptions);
  case 'fminsearch'
    [Par_est, F_val, ExitFlag] = fminsearch(objfun,ParVar,OptimOptions);
  case 'patternsearch'
    [Par_est, F_val, ExitFlag] = patternsearch(objfun,ParVar,[],[],[],[],LB,UB,confun,OptimOptions);
  otherwise
    error('Optimization Method is not well defined');    
end     
warning('on','all');
disp(['EXITFLAG: ',num2str(ExitFlag)])

t          = EstimOptions{1};
Z          = EstimOptions{2};
Pf0        = EstimOptions{4};
FileNames  = EstimOptions{6};
LKWindow   = EstimOptions{7};
mpts       = EstimOptions{16};
AbsTol     = EstimOptions{18};
RelTol     = EstimOptions{19};

Xf0_est=Xf00;
Theta_est=Theta0;  
switch EstimOptions{13},
  case 0,
     NoZeroXf00   = intersect(find(abs(Xf00)>eps),Xf0Index);
     ZeroXf00     = intersect(find(abs(Xf00)<eps),Xf0Index);
     NoZeroTheta0 = intersect(find(abs(Theta0)>eps),ThetaIndex);
     ZeroTheta0   = intersect(find(abs(Theta0)<eps),ThetaIndex);
     AllPar=[Xf0_est' Theta_est];
     
     n_v_e=length(Xf00);
     AllPar([NoZeroXf00 n_v_e+NoZeroTheta0]) = [Xf00(NoZeroXf00)' Theta0(NoZeroTheta0)].* Par_est(NoZeroParVar);
     AllPar([ZeroXf00 ZeroTheta0])     = Par_est(ZeroParVar) - 1;  
     Xf0_est   = AllPar(1:n_v_e)';
     Theta_est = AllPar(1+n_v_e:end);
  case 1, 
     NoZeroTheta0 = intersect(find(abs(Theta0)>eps),ThetaIndex);
     ZeroTheta0   = intersect(find(abs(Theta0)<eps),ThetaIndex);
     Theta_est(NoZeroTheta0) = Theta0(NoZeroTheta0) .*  Par_est(NoZeroParVar);
     Theta_est(ZeroTheta0)   = Par_est(ZeroParVar) - 1;
  case 2,
     NoZeroXf00 = intersect(find(abs(Xf00)>eps),Xf0Index);
     ZeroXf00   = intersect(find(abs(Xf00)<eps),Xf0Index);
     Xf0_est(NoZeroXf00) = Xf00(NoZeroXf00) .* Par_est(NoZeroParVar)';
     Xf0_est(ZeroXf00)   = Par_est(ZeroParVar)' - 1;
end
  
[Xff,Pff,Xfp,Pfp,FInn,LogLK_est,PFInn,mpts_out]=LLfilter(t,Z,Xf0_est,Pf0,Theta_est,FileNames,LKWindow,mpts,AbsTol,RelTol);



%-------------------------------------------

function option = EstimChoise(ThetaIndex,Xf0Index)

if (~isempty(ThetaIndex)) & (isempty(Xf0Index)) , 
   option=1;
else  
   if (isempty(ThetaIndex)) & (~isempty(Xf0Index)) ,
      option=2;
   else
      option=0;
   end;
end;   





