function z = EulerScheme(f,g,x0,t,theta)
%   Euler scheme for the numerical integrator of the stochastic differential
% equations with multiplicative noise: 
%	  dx = f(t,x,theta) dt + g(t,x,theta) dw
%
% Notation
%   n_v_e: number of state variables x 
%
% Imputs:
%   x0:     initial value of x at t0  (n_v_e x 1) 
%   f:      string with the m file name of f
%   g:      string with the m file name of g, 
%   theta:  parameters of the differential equation
%
% Imputs of the above functions 
%   f(t,x,theta)           g(t,x,theta)
%
% Outputs:
%   z:   numerical solucion obtained by the LL scheme (n_v_e x n_m)

n_v_e = size(x0,1);
n_m = length(t);
z = zeros(n_v_e,n_m);
deltat=[0, diff(t)]';

x = x0;
te = t(1);

eta = randn(n_m,n_v_e);  
for i=1:n_v_e,
  eta(:,i) = sqrt(deltat) .* eta(:,i);    % system noise generation 
end
w = cumsum([zeros(1,n_v_e);eta]);
dw=diff(w)';

z(:,1) = x;
for tt = 1:n_m-1
   fe = feval(f,te,x,theta);
   ge = feval(g,te,x,theta);
      
   m=size(ge,2);
   ruido=zeros(n_v_e,1);
   for k=1:m 
      ruido=ruido + ge{k}.*dw(k,tt);
   end;   
   te = t(tt+1);
   x = x + fe.*deltat(tt+1) + ruido;

   z(:,tt+1) = x;
   
end% for en tt

