% Script for computating the approximate innovation estimators and approximate fitting-filter 
% proposed in Jimenez & Ozaki (2002) for diffusion processes: 
%     dx = f(t,x,theta) dt + g(t,x,theta) dw, 
% given a set of discrete observations of some components of x. 

% These estimators for the uknown parameters theta and unobserved components of the process x are computed following 
% the heuristic introduced in Ozaki et. al (2000). This heuristic essential has three steps: a) computation of the 
% parameters theta; b) computation of the unobserved initial state of x; c) computation of the approxiamte-fitting 
% filter.
% For actual data a carefull selection of the aditional parameters: firstpoint, amplitude as well the options for the 
% optimization algorithm "fmincon" is very important to obtain successful results. In the same way an adequate 
% selection of the initial values for the parameters theta (theta0) and initial state (Xf0, Pf0) are crucial. 
% Thus, import modification must be done in the scripts for the generation and test that run before the present script.  
% Typically, before to run the present script many simulation with the scripts for generation and test must be done 
% to select adecuate values for the variables theta0, X0 and V0. 

% References:
%   -  "An approximate innovation method for the estimation of diffusion processes from discrete data", 
%       Jimenez J.C. & Ozaki, J. Time Series Analysis 
%   -  "Role of the Likelihood Function in the Estimation of Chaos Models", 
%       Ozaki T., Jimenez J.C. & Haggab-Ozaki V., 
%       J. of Time Series Analysis (2000), Vol. 21, 363-387. 


echo off

global deltat deltat_min n_v_e n_m n_obs t0 tTotal X0 Xf0 Pf0 mpts
global Theta FileNames
global gct             % variable to be used in confun.m

%Definicion de todos las constantes
const_cox;

load data 
Z=datos;

%*********************  Estimation of the unknown model parameters (theta)
ch='P';
while ~(strcmp(ch,'Y') | strcmp(ch,'N'))
  ch=upper(input('Optimization with a priory statistics on the innovations? [y/n]= ','s'));
end;
if ch=='Y', 
  firstpoint=1;     % starting point for the computation of the innovations statistics
  lastpoint=n_m;    % last point for the computation of the innovations statistics
  nut=nu(:,firstpoint:lastpoint)';
  amplitude=1;     
  nu_stast=amplitude.*[abs(mean(nut)) diag(cov(nut))'];
else  
  nu_stast=[];
end

LKWindow=[t0 tTotal];
Xf0Index=[];
%LB=[];
%UB=[];
LB=[0 -3  0  -3  0 ];
UB=[2  0 0.5  0 0.5];
ThetaIndex=[1 2 3 4 5];
Theta0=Theta;                                     % convergence ok
%Theta0(1:5)=[0.9 -1.6 0.12 -1.4 0.015];           % so so 
%Theta0(1:5)=[0.5 -2 0.2 -0.4 0.04];               % not ok (to far for this large system noise)

%Pf0 should be convenient adjuted if Xf0 are going to be estimated
OptimOptions = optimset('fmincon');
OptimOptions = optimset(OptimOptions,'LargeScale','off','Display','iter','MaxFunEval',1e3);
EstimOptions = estimset(t,Z,Xf0,Pf0,Theta0,FileNames,LKWindow,mpts,nu_stast,LB,UB,OptimOptions);
[Xff,Pff,Xfp,Pfp,nu,LogLK_est,Xf0_Est,Theta_Est]=InnovEstimators(Xf0Index,ThetaIndex,EstimOptions);

% salvar estimacion

disp(['   '])
disp([' Log LK for Theta_Est= ' num2str(-LogLK_est)])
disp(['   '])
disp(['        theta          theta0          theta_est'])
[Theta(ThetaIndex)' Theta0(ThetaIndex)' Theta_Est(ThetaIndex)']

figure;
subplot(1,2,1)
plot(nu);
subplot(1,2,2);
hist(nu,20);
drawnow;


%*********************  Estimation of the unobserved initial state (Xf0)
firstpoint=1;     % starting point for the computation of the innovations statistics
lastpoint=n_m;    % last point for the computation of the innovations statistics
nut=nu(firstpoint:lastpoint);
amplitude=1.2;     
nu_stast=amplitude.*[abs(mean(nut)) cov(nut)];

LKWindow=[t0 tTotal];
%LB=[];
%UB=[];
LB=[0];
UB=[1];
Xf0Index=[1];
ThetaIndex=[];
Xf00=Xf0;       % convergence ok

%Pf0 should be convenient adjuted after the estimation of theta
OptimOptions = optimset('fmincon');
OptimOptions = optimset(OptimOptions,'LargeScale','off','Display','iter','MaxFunEval',1e3);
EstimOptions = estimset(t,Z,Xf00,Pf0,Theta_Est,FileNames,LKWindow,mpts,nu_stast,LB,UB,OptimOptions);
[Xff,Pff,Xfp,Pfp,nu,LogLK_est,Xf0_Est,Theta_Est]=InnovEstimators(Xf0Index,ThetaIndex,EstimOptions);

% salvar estimacion

disp(['   '])
disp([' Log LK for Theta_Est= ' num2str(-LogLK_est)])
disp(['   '])
disp(['       Xf0           Xf00         Xf0_est'])
[Xf0(Xf0Index)' Xf00(Xf0Index)' Xf0_Est(Xf0Index)']

figure;
subplot(1,2,1)
plot(nu);
subplot(1,2,2);
hist(nu,20);
'innovation mean'
mean(nu)
'innovation variance'
cov(nu)


%el tercer paso fu borrado


%****************  plot of exact filters vs. approximate-fitting filter (and their variances)
ps=zeros(n_m,1);
t(1)=0;
ms(1)=Xf0(1);
ps(1)=Pf0(1,1);
for i=2:n_m
   t(i)=t(i-1)+deltat;
   ms(i)=ms(i-1).*exp(Theta(2)*deltat)+Theta(1).*(exp(Theta(2).*deltat)-1)./Theta(2);
   temp = (Theta(1)+2.*Theta(2).*ms(1)).*exp(2.*Theta(2).*t(i)) ... 
      -2.*(Theta(1)+Theta(2).*ms(1)).*exp(Theta(2).*t(i)) + Theta(1);
   ps(i)= (temp./2).*(Theta(3)/Theta(2)).^2;
end;
figure
subplot(2,1,1);
plot(t,ms,t,Xff(1,:));
subplot(2,1,2);
plot(t,ps,t,Pff(:,1,1));
drawnow;

