function [Xf,Pf,Xp,Pp,Inn,LogLK,PInn,mpts]=LLfilter_Adapt(t,Z,Xf0,Pf0,theta,FileNames,LKWindow,AbsTol,RelTol)
%    Local Linearization Filter for the estimation of the unobserved componet x of the 
% continuous model
%        dx = f(t,x,theta) dt + g(t,x,theta) dw, 
% with dicrete observation equation 
%         Z = h(t,x,theta) + p(t,x,theta)n + e, 
% given a set of discrete observation of Z. Xf0, Pf0 and theta are also known and given in advance.
%
% Notation
%   n_m: number of time-instants where Z is observed
%   n_obs: number of observation equations
%   n_v_e: number of state variables x 
%   n_theta: number of parameters theta 
%
% Input:
%   t: instants of times where at least one component of Z is observed  (1 x n_m)          
%   Z: observed sample                        (n_obs x n_m)   Set Z(k,t)=inf when data Z(k,t) is missing
%   Xf0: initial filter value                 (n_v_e x 1)
%   Pf0: initial filter variance value        (n_v_e x n_v_e)
%   theta: State Space Model parameters       (1 x n_theta)
%   FileNames: cell variable with the name of all the functions that define State Space Model
%   LKWindow: interval of time where the likelihood function will be computed  (1 x 2)
%   AbsTol: Absolute Tolerance (1 x 2), 1: for the first moment, 2: for the second moment
%   RelTol: Relative Tolerance (1 x 2), 1: for the first moment, 2: for the second moment
%
%   FileNames={f,g,Jf,Jg,Hf,Hg,ft,gt,h,Jh,p,Jp,See,Lee}   
%       f string with the m file name of drift coefficient f
%       g string with the m file name of diffusion coefficient g
%      Jf string with the m file name of the Jacobian of f 
%      Jg string with the m file name of the Jacobian of g
%      Hf string with the m file name of the Hessian of f 
%      Hg string with the m file name of the Hessian of g
%      ft string with the m file name of derivative of f respect to t
%      gt string with the m file name of derivative of g respect to t
%       h string with the m file name of h 
%       p string with the m file name of p 
%      Jh string with the m file name of the Jacobian of h 
%      Jp string with the m file name of the Jacobian of p 
%     See string with the m file name of the covariance matrix of the additive noise e 
%     Lee string with the m file name of the covariance matrix of the multiplicative noise n 
%
% OutPuts:
%   Xf : LL filter  
%   Pf : LL filter variance 
%   Xp : LL prediction 
%   Pp : LL prediction variance  
%   Inn: Innovation
%   LogLK:  -2log(Likelihood)
%   PInn: Innovation variance
%   mpts: number of misssing points to compute the prediction between two observations (2 x n_m)
%         1: accepted, 2: failed

%  References
%   - "Local Linearization filters for non-linear continuous-discrete state 
%     space models with multiplicative noise", Jimenez J.C. & Ozaki
%     Int. J. Control (2003), 76, 1159-1170. 
%   - "Linear estimation of continuous-discrete linear state space models 
%     with multiplicative noise",  Jimenez J.C. & Ozaki T.
%     Systems & Control Letters (2002), 47, 91-101.

n_m=size(t,2);        % number of instants of times
n_v_e=size(Xf0,1);    % number of state variables
nobs=size(Z,1);       % number of observed variables
if strcmp(FileNames{6},'zero'),  % pp=1/(1+order)   with order= 1 or 2 
    pp=0.5; 
else
    pp=1/3;
end

Xf=zeros(n_v_e,n_m);
Xp=zeros(n_v_e,n_m);
Pf=zeros(n_m,n_v_e,n_v_e);
Pp=zeros(n_m,n_v_e,n_v_e);
Inn=zeros(nobs,n_m);
PInn=zeros(n_m,nobs,nobs);
mpts=zeros(2,n_m);

Zt=Xf0;
Vt=Pf0;
Xf(:,1)=Zt;
Xp(:,1)=Zt;
Pf(1,:,:) = Pf0;
Pp(1,:,:) = Pf0;

LogLK = 0.0;
[Prediction, Filter] = ModObsEqType(FileNames);
hh=initial_step(t(1),Zt,Vt + Zt*Zt',theta,FileNames,AbsTol,RelTol,pp);
for k= 2:n_m
   
  Ztdt=Zt;
  Pt  =Vt + Zt*Zt';
  
  ti=t(k-1);
%  hh=initial_step(ti,Ztdt,Pt,theta,FileNames,AbsTol,RelTol,pp);

  while t(k)-ti>eps,
     TT=0.5.*(t(k)-ti);
     if hh>TT, hh=TT; end;
     hh2=2.*hh;
     ti_hh =ti+hh;
     ti_hh2=ti+hh2;
     [Ztdt1, Pt1, Ztdt2, Pt2] = feval(Prediction,ti_hh,ti,Ztdt,Pt,theta,FileNames);  
     [Ztdt3, Pt3] = feval(Prediction,ti_hh2,ti_hh,Ztdt1,Pt1,theta,FileNames);  
     errorZ=error_step(Ztdt,Ztdt3,Ztdt2,n_v_e,AbsTol(1),RelTol(1));
     errorP=error_step(vec(Pt),vec(Pt3),vec(Pt2),sqr(n_v_e),sqr(AbsTol(2)),RelTol(2));
     if (errorZ<=1) && (errorP<=1),
        ti=ti_hh2;
        Ztdt=Ztdt3;
        Pt=Pt3;
        hhZ=change_step_no_reject(hh,errorZ,pp);
        hhP=change_step_no_reject(hh,errorP,pp);
        mpts(1,k)=mpts(1,k)+1;
     else
        if errorZ>1,
           hhZ=change_step_reject(hh,errorZ,pp);
        else
           hhZ=change_step_no_reject(hh,errorZ,pp);         
        end
        if errorP>1,
           hhP=change_step_reject(hh,errorP,pp);
        else   
           hhP=change_step_no_reject(hh,errorP,pp);
        end
        mpts(2,k)=mpts(2,k)+1;
     end   
     hh=max(eps,min([hhZ hhP]));
  end      
  Pt = Pt - Ztdt*Ztdt';
  
%   hh  =(t(k)-t(k-1))/(mpts(k-1)+1);
%   for j=1:(mpts(k-1)+1),                                                   % LL prediction
%     [Ztdt, Pt] = feval(Prediction,t(k-1)+j*hh,t(k-1)+(j-1)*hh,Ztdt,Pt,theta,FileNames);         
%   end;
  
  [Zt, Vt, nu, Sv] = feval(Filter,t(k),Ztdt,Pt,theta,Z(:,k),FileNames);    % LL filter & innovation 
  
  if (LKWindow(1)<=t(k)) && (t(k)<=LKWindow(2)),
     LogLK = LogLK + log(det(Sv)) + nu' /Sv * nu;                          % quasi-likelihood
  end
    
  Xp(:,k)   = Ztdt;    % LL prediction
  Pp(k,:,:) = Pt;      % LL prediction variance
  Xf(:,k)   = Zt;      % LL filter 
  Pf(k,:,:) = Vt;      % LL filter variance
  
  obs= find(Z(:,k)~=inf);
  Inn(obs,k)  = nu;      % innovation
  PInn(k,obs,obs) = Sv;      % innovation variance
   
end

LogLK = LogLK + log(2.*pi).*n_m;


%----------------------------

function b = bxt(x,g,Jg)
m=size(Jg,2);
for i=1:m
   b{i}=g{i}-Jg{i}*x;
end;   


%----------------------------

function v = vec(x);

[n m]=size(x);
v=reshape(x,n*m,1);


%----------------------------
% Defines the type of Model and Observation Equations

function [Prediction, Filter] = ModObsEqType(FileNames)

Jg=FileNames(4);
ft=FileNames(7);
gt=FileNames(8);
p=FileNames(11);

if strcmp(Jg,'zero'),
   if ( strcmp(ft,'zero') & strcmp(gt,'zero') ),  
      Prediction=@Prediction_Auto_Add; 
   else
      Prediction=@Prediction_NoAuto_Add;
   end;
else
   if ( strcmp(ft,'zero') & strcmp(gt,'zero') ),  
      Prediction=@Prediction_Auto_Mult; 
   else
      Prediction=@Prediction_NoAuto_Mult;
   end;
end;

if strcmp(p,'zero'),
   Filter=@Filter_Add;
else
   Filter=@Filter_Mult;
end;   


%----------------------------
% LL prediction estimates for Autonomous Equations with Multiplicative Noise

function [Ztdt, Pt, Ztdt1, Pt1] = Prediction_Auto_Mult(tt,tt0,Zt,Vt,theta,FileNames)

fe  = feval(FileNames{1},tt0,Zt,theta);
ge  = feval(FileNames{2},tt0,Zt,theta);
Je  = feval(FileNames{3},tt0,Zt,theta);
Jge = feval(FileNames{4},tt0,Zt,theta);

be  = bxt(Zt,ge,Jge);
 
if nargout>2,
  [Ztdt, Pt, Ztdt1, Pt1] = LLpredictionAuto(tt-tt0,Je,fe,Jge,be,Zt,Vt); 
else
  [Ztdt, Pt] = LLpredictionAuto(tt-tt0,Je,fe,Jge,be,Zt,Vt); 
end
 

%----------------------------
% LL prediction estimates for NoAutonomous Equations with Multiplicative Noise

function [Ztdt, Pt, Ztdt1, Pt1] = Prediction_NoAuto_Mult(tt,tt0,Zt,Vt,theta,FileNames)

fe  = feval(FileNames{1},tt0,Zt,theta);
ge  = feval(FileNames{2},tt0,Zt,theta);
Je  = feval(FileNames{3},tt0,Zt,theta);
Jge = feval(FileNames{4},tt0,Zt,theta);
fte  = feval(FileNames{7},tt0,Zt,theta);
gte  = feval(FileNames{8},tt0,Zt,theta);

be  = bxt(Zt,ge,Jge);

if nargout>2,
   [Ztdt, Pt, Ztdt1, Pt1] = LLpredictionNoAuto(tt-tt0,Je,fe,Jge,be,Zt,Vt,fte,gte); 
else
   [Ztdt, Pt] = LLpredictionNoAuto(tt-tt0,Je,fe,Jge,be,Zt,Vt,fte,gte); 
end


%----------------------------
% LL prediction estimates for NoAutonomous Equations with Additive Noise

function [Ztdt, Pt, Ztdt1, Pt1] = Prediction_NoAuto_Add(tt,tt0,Zt,Vt,theta,FileNames)

fe  = feval(FileNames{1},tt0,Zt,theta);
ge  = feval(FileNames{2},tt0,Zt,theta);
Je  = feval(FileNames{3},tt0,Zt,theta);
fte  = feval(FileNames{7},tt0,Zt,theta);
gte  = feval(FileNames{8},tt0,Zt,theta);

if nargout>2,
   [Ztdt, Pt, Ztdt1, Pt1] = LLpredictionNoAutoAdd(tt-tt0,Je,fe,ge,Zt,Vt,fte,gte); 
else
   [Ztdt, Pt] = LLpredictionNoAutoAdd(tt-tt0,Je,fe,ge,Zt,Vt,fte,gte); 
end


%----------------------------
% LL prediction estimates for Autonomous Equations with Additive Noise

function [Ztdt, Pt, Ztdt1, Pt1] = Prediction_Auto_Add(tt,tt0,Zt,Vt,theta,FileNames)

fe  = feval(FileNames{1},tt0,Zt,theta);
ge  = feval(FileNames{2},tt0,Zt,theta);
Je  = feval(FileNames{3},tt0,Zt,theta);

if nargout>2,
   [Ztdt, Pt, Ztdt1, Pt1] = LLpredictionAutoAdd(tt-tt0,Je,fe,ge,Zt,Vt); 
else
   [Ztdt, Pt] = LLpredictionAutoAdd(tt-tt0,Je,fe,ge,Zt,Vt); 
end


%----------------------------
% LL filter estimates for Observations with Additive Noise

function [Zt, Vt, nu, Sv] = Filter_Add(tt,Ztdt,Pt,theta,X,FileNames)

h  =FileNames{9};
Jh =FileNames{10};
See=feval(FileNames{13},theta,X);

obs= find(X(:)~=inf);
nu = X(obs) - feval(h,tt,Ztdt,theta,X);         %  innovation

Jhe = feval(Jh,tt,Ztdt,theta,X);     
%Sv   = Jhe*Pt*Jhe' + See(obs,obs);               
Sv   = Jhe*Pt*Jhe' + See;               
Kt = Pt*Jhe' / Sv;                      %(4)    LL filter gain
Zt = Ztdt + Kt * nu;                    %(5)    LL filter
Vt = Pt - Kt * Jhe*Pt;                  %(6)    LL filter variance


%----------------------------
% LL filter estimates for Observations with Multiplicative Noise

function [Zt, Vt, nu, Sv] = Filter_Mult(tt,Ztdt,Pt,theta,X,FileNames)

hh =FileNames{9};
Jh =FileNames{10};
p  =FileNames{11};
Jp =FileNames{12};
See=feval(FileNames{13},theta,X);
L  =feval(FileNames{14},theta,X);

obs= find(X(:)~=inf);
nu = X(obs) - feval(hh,tt,Ztdt,theta,X);         %  innovation

Jhe = feval(Jh,tt,Ztdt,theta,X);     
pe  = feval(p,tt,Ztdt,theta,X);
Jpe = feval(Jp,tt,Ztdt,theta,X);
n=size(pe,2);
p_i=See;
for j=1:n
   th{j} = pe{j}-Jpe{j}*Ztdt;
   p_i   = p_i + L(j)*th{j}*th{j}';
   th{j} = L(j)*th{j};
end;
Sv   = Jhe*Pt*Jhe' + p_i;                  % es parte de (4)
for j=1:n
   temp=Jpe{j}*Ztdt*th{j}';
   Sv = Sv + L(j)*Jpe{j}*(Pt+Ztdt*Ztdt')*Jpe{j}' + temp + temp';     % innovation variance
end;   
Kt = Pt*Jhe' / Sv;                      %(4)    LL filter gain
Zt = Ztdt + Kt * nu;                    %(5)    LL filter
Vt = Pt - Kt * Jhe*Pt;                  %(6)    LL filter variance

function  hh=initial_step(t0,m0,P0,theta,FileNames,AbsTol,RelTol,pp)

d=size(m0,1);
f0=feval(FileNames{1},t0,m0,theta);
A=feval(FileNames{3},t0,m0,theta); 
if strcmp(FileNames{7},'zero'),
  a1=zeros(d,1);
else
  a1=feval(FileNames{7},t0,m0,theta);
end
hh_m=starting_step(f0,A,a1,m0,d,AbsTol(1),RelTol(1),pp);

d2=d*d;
I_d=eye(d);
bbT=zeros(d,d);
b1bT=zeros(d,d);
bb1T=zeros(d,d);
BK{1}=zeros(d2,d2);
BK{2}=zeros(d2,d);
BK{3}=zeros(d2,d);
BK{4}=zeros(d2,d);
BK{5}=zeros(d2,d);
g0  = feval(FileNames{2},t0,m0,theta);
m=size(g0,2);
if strcmp(FileNames{4},'zero'),    %B=0
   if strcmp(FileNames{8},'zero'), %b1=0;
      for i=1:m 
        bbT = bbT + g0{i}*g0{i}';
      end
   else
      b1  = feval(FileNames{8},t0,m0,theta);
      for i=1:m 
        bbT = bbT + g0{i}*g0{i}';
        b1bT = b1bT + b1{i}*g0{i}';
        bb1T = bb1T + g0{i}*b1{i}';
      end
   end
else
   B = feval(FileNames{4},t0,m0,theta);
   b  = bxt(m0,g0,B);
   if strcmp(FileNames{8},'zero'), %b1=0;
     for i=1:m 
       bbT = bbT + b{i}*b{i}';
       BK{1}= BK{1} + kron(B{i},B{i});
       BK{2}= BK{2} + kron(b{i},B{i});
       BK{3}= BK{3} + kron(B{i},b{i});
     end
   else
     b1  = feval(FileNames{8},t0,m0,theta);
     for i=1:m 
       bbT = bbT + b{i}*b{i}';
       BK{1}= BK{1} + kron(B{i},B{i});
       BK{2}= BK{2} + kron(b{i},B{i});
       BK{3}= BK{3} + kron(B{i},b{i});
   
       BK{4}= BK{4} + kron(b1{i},B{i});
       BK{5}= BK{5} + kron(B{i},b1{i});
       b1bT = b1bT + b1{i}*b{i}';
       bb1T = bb1T + b{i}*b1{i}';
     end
   end
end
a0=f0-A*m0;
AE = kron(I_d,A) + kron(A,I_d) + BK{1};
beta1 = vec(bbT);
beta2 = vec(b1bT + bb1T);
beta4 = kron(I_d,a0) + kron(a0,I_d) + BK{2} + BK{3};
beta5 = kron(I_d,a1) + kron(a1,I_d) + BK{4} + BK{5};
F0=AE*vec(P0)+beta1+beta4*m0;
Ft0=beta2+2*beta5*m0+beta4*f0;
hh_P=starting_step(F0,AE,Ft0,vec(P0),d2,AbsTol(2),RelTol(2),pp);

hh = max(eps,min([hh_m; hh_P]));



