function [Xf,Pf,Xp,Pp,Inn,LogLK,PInn,mpts]=LLfilter_Mpts(t,Z,Xf0,Pf0,theta,FileNames,LKWindow,mpts)
%    Local Linearization Filter for the estimation of the unobserved componet x of the 
% continuous model
%        dx = f(t,x,theta) dt + g(t,x,theta) dw, 
% with dicrete observation equation 
%         Z = h(t,x,theta) + p(t,x,theta)n + e, 
% given a set of discrete observation of Z. Xf0, Pf0 and theta are also known and given in advance.
%
% Notation
%   n_m: number of time-instants where Z is observed
%   n_obs: number of observation equations
%   n_v_e: number of state variables x 
%   n_theta: number of parameters theta 
%
% Input:
%   t: instants of times where at least one component of Z is observed  (1 x n_m)          
%   Z: observed sample                        (n_obs x n_m)   Set Z(k,t)=inf when data Z(k,t) is missing
%   Xf0: initial filter value                 (n_v_e x 1)
%   Pf0: initial filter variance value        (n_v_e x n_v_e)
%   theta: State Space Model parameters       (1 x n_theta)
%   FileNames: cell variable with the name of all the functions that define State Space Model
%   LKWindow: interval of time where the likelihood function will be computed  (1 x 2)
%   mpts: number of misssing points to compute the prediction between two observations (1 x n_m)
%
%   FileNames={f,g,Jf,Jg,Hf,Hg,ft,gt,h,Jh,p,Jp,See,Lee}   
%       f string with the m file name of drift coefficient f
%       g string with the m file name of diffusion coefficient g
%      Jf string with the m file name of the Jacobian of f 
%      Jg string with the m file name of the Jacobian of g
%      Hf string with the m file name of the Hessian of f 
%      Hg string with the m file name of the Hessian of g
%      ft string with the m file name of derivative of f respect to t
%      gt string with the m file name of derivative of g respect to t
%       h string with the m file name of h 
%       p string with the m file name of p 
%      Jh string with the m file name of the Jacobian of h 
%      Jp string with the m file name of the Jacobian of p 
%     See string with the m file name of the covariance matrix of the additive noise e 
%     Lee string with the m file name of the covariance matrix of the multiplicative noise n 
%
% OutPuts:
%   Xf : LL filter  
%   Pf : LL filter variance 
%   Xp : LL prediction 
%   Pp : LL prediction variance  
%   Inn: Innovation
%   LogLK:  -2log(Likelihood)
%   PInn: Innovation variance

%  References
%   - "Local Linearization filters for non-linear continuous-discrete state 
%     space models with multiplicative noise", Jimenez J.C. & Ozaki
%     Int. J. Control (2003), 76, 1159-1170. 
%   - "Linear estimation of continuous-discrete linear state space models 
%     with multiplicative noise",  Jimenez J.C. & Ozaki T.
%     Systems & Control Letters (2002), 47, 91-101.

n_m=size(t,2);        % number of instants of times
n_v_e=size(Xf0,1);    % number of state variables
nobs=size(Z,1);       % number of observed variables

Xf=zeros(n_v_e,n_m);
Xp=zeros(n_v_e,n_m);
Pf=zeros(n_m,n_v_e,n_v_e);
Pp=zeros(n_m,n_v_e,n_v_e);
Inn=zeros(nobs,n_m);
PInn=zeros(n_m,nobs,nobs);

Zt=Xf0;
Vt=Pf0;
Xf(:,1)=Zt;
Xp(:,1)=Zt;
Pf(1,:,:) = Pf0;
Pp(1,:,:) = Pf0;

LogLK = 0.0;
[Prediction, Filter] = ModObsEqType(FileNames);
for k= 2:n_m
   
  Ztdt=Zt;
  Pt  =Vt + Zt*Zt';
  
  hh  =(t(k)-t(k-1))/(mpts(k-1)+1);
  for j=1:(mpts(k-1)+1),                                                   % LL prediction
    [Ztdt, Pt] = feval(Prediction,t(k-1)+j*hh,t(k-1)+(j-1)*hh,Ztdt,Pt,theta,FileNames);         
  end;
  Pt = Pt - Ztdt*Ztdt';
    
  [Zt, Vt, nu, Sv] = feval(Filter,t(k),Ztdt,Pt,theta,Z(:,k),FileNames);    % LL filter & innovation 
   
  if (LKWindow(1)<=t(k)) & (t(k)<=LKWindow(2)),
     LogLK = LogLK + log(det(Sv)) + nu' /Sv * nu;                          % quasi-likelihood
  end
  
  Xp(:,k)   = Ztdt;    % LL prediction
  Pp(k,:,:) = Pt;      % LL prediction variance
  Xf(:,k)   = Zt;      % LL filter 
  Pf(k,:,:) = Vt;      % LL filter variance
  
  obs= find(Z(:,k)~=inf);
  Inn(obs,k)  = nu;      % innovation
  PInn(k,obs,obs) = Sv;      % innovation variance
   
end

LogLK = LogLK + log(2.*pi).*n_m;


%----------------------------

function b = bxt(x,g,Jg)
m=size(Jg,2);
for i=1:m
   b{i}=g{i}-Jg{i}*x;
end;   


%----------------------------
% Defines the type of Model and Observation Equations

function [Prediction, Filter] = ModObsEqType(FileNames)

Jg=FileNames(4);
ft=FileNames(7);
gt=FileNames(8);
p=FileNames(11);

if strcmp(Jg,'zero'),
   if ( strcmp(ft,'zero') & strcmp(gt,'zero') ),  
      Prediction=@Prediction_Auto_Add; 
   else
      Prediction=@Prediction_NoAuto_Add;
   end;
else
   if ( strcmp(ft,'zero') & strcmp(gt,'zero') ),  
      Prediction=@Prediction_Auto_Mult; 
   else
      Prediction=@Prediction_NoAuto_Mult;
   end;
end;

if strcmp(p,'zero'),
   Filter=@Filter_Add;
else
   Filter=@Filter_Mult;
end;   


%----------------------------
% LL prediction estimates for Autonomous Equations with Multiplicative Noise

function [Ztdt, Pt] = Prediction_Auto_Mult(tt,tt0,Zt,Vt,theta,FileNames)

fe  = feval(FileNames{1},tt0,Zt,theta);
ge  = feval(FileNames{2},tt0,Zt,theta);
Je  = feval(FileNames{3},tt0,Zt,theta);
Jge = feval(FileNames{4},tt0,Zt,theta);

be  = bxt(Zt,ge,Jge);
 
[Ztdt, Pt] = LLpredictionAuto(tt-tt0,Je,fe,Jge,be,Zt,Vt); 
 

%----------------------------
% LL prediction estimates for NoAutonomous Equations with Multiplicative Noise

function [Ztdt, Pt] = Prediction_NoAuto_Mult(tt,tt0,Zt,Vt,theta,FileNames)

fe  = feval(FileNames{1},tt0,Zt,theta);
ge  = feval(FileNames{2},tt0,Zt,theta);
Je  = feval(FileNames{3},tt0,Zt,theta);
Jge = feval(FileNames{4},tt0,Zt,theta);
fte  = feval(FileNames{7},tt0,Zt,theta);
gte  = feval(FileNames{8},tt0,Zt,theta);

be  = bxt(Zt,ge,Jge);
 
[Ztdt, Pt] = LLpredictionNoAuto(tt-tt0,Je,fe,Jge,be,Zt,Vt,fte,gte); 


%----------------------------
% LL prediction estimates for NoAutonomous Equations with Additive Noise

function [Ztdt, Pt] = Prediction_NoAuto_Add(tt,tt0,Zt,Vt,theta,FileNames)

fe  = feval(FileNames{1},tt0,Zt,theta);
ge  = feval(FileNames{2},tt0,Zt,theta);
Je  = feval(FileNames{3},tt0,Zt,theta);
fte  = feval(FileNames{7},tt0,Zt,theta);
gte  = feval(FileNames{8},tt0,Zt,theta);

[Ztdt, Pt] = LLpredictionNoAutoAdd(tt-tt0,Je,fe,ge,Zt,Vt,fte,gte); 


%----------------------------
% LL prediction estimates for Autonomous Equations with Additive Noise

function [Ztdt, Pt] = Prediction_Auto_Add(tt,tt0,Zt,Vt,theta,FileNames)

fe  = feval(FileNames{1},tt0,Zt,theta);
ge  = feval(FileNames{2},tt0,Zt,theta);
Je  = feval(FileNames{3},tt0,Zt,theta);

[Ztdt, Pt] = LLpredictionAutoAdd(tt-tt0,Je,fe,ge,Zt,Vt); 


%----------------------------
% LL filter estimates for Observations with Additive Noise

function [Zt, Vt, nu, Sv] = Filter_Add(tt,Ztdt,Pt,theta,X,FileNames);

h  =FileNames{9};
Jh =FileNames{10};
See=feval(FileNames{13},theta,X);

obs= find(X(:)~=inf);
nu = X(obs) - feval(h,tt,Ztdt,theta,X);         %  innovation

Jhe = feval(Jh,tt,Ztdt,theta,X);     
%Sv   = Jhe*Pt*Jhe' + See(obs,obs);               
Sv   = Jhe*Pt*Jhe' + See;               
Kt = Pt*Jhe' / Sv;                      %(4)    LL filter gain
Zt = Ztdt + Kt * nu;                    %(5)    LL filter
Vt = Pt - Kt * Jhe*Pt;                  %(6)    LL filter variance


%----------------------------
% LL filter estimates for Observations with Multiplicative Noise

function [Zt, Vt, nu, Sv] = Filter_Mult(tt,Ztdt,Pt,theta,X,FileNames);

hh =FileNames{9};
Jh =FileNames{10};
p  =FileNames{11};
Jp =FileNames{12};
See=feval(FileNames{13},theta,X);
L  =feval(FileNames{14},theta,X);

obs= find(X(:)~=inf);
nu = X(obs) - feval(hh,tt,Ztdt,theta,X);         %  innovation

Jhe = feval(Jh,tt,Ztdt,theta,X);     
pe  = feval(p,tt,Ztdt,theta,X);
Jpe = feval(Jp,tt,Ztdt,theta,X);
n=size(pe,2);
p_i=See;
for j=1:n
   th{j} = pe{j}-Jpe{j}*Ztdt;
   p_i   = p_i + L(j)*th{j}*th{j}';
   th{j} = L(j)*th{j};
end;
Sv   = Jhe*Pt*Jhe' + p_i;                  % es parte de (4)
for j=1:n
   temp=Jpe{j}*Ztdt*th{j}';
   Sv = Sv + L(j)*Jpe{j}*(Pt+Ztdt*Ztdt')*Jpe{j}' + temp + temp';     % innovation variance
end;   
Kt = Pt*Jhe' / Sv;                      %(4)    LL filter gain
Zt = Ztdt + Kt * nu;                    %(5)    LL filter
Vt = Pt - Kt * Jhe*Pt;                  %(6)    LL filter variance
